# pylint: skip-file
import unittest

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine

from unimatrix.ext.orm.conf import load_config


class EngineConfigurationTestCase(unittest.TestCase):

    def setUp(self):
        self.connections = load_config({
            'DB_ENGINE': 'sqlite',
            'DB_NAME': ":memory:"
        })
        self.connection = self.connections['self']

    def test_create_engine_from_config(self):
        engine = self.connection.as_engine(create_engine)
        self.assertIsInstance(engine, Engine)



"""Provides functions to configure the database connection(s)."""
import copy
import os

import unimatrix.lib.rdbms


class DatabaseConnection(dict):
    """Represents a database connection configuration."""

    def as_dsn(self, use_async=False):
        """Returns the connection credentials as a data source name."""
        if self['engine'] and self.get('name') == ':memory:':
            return "sqlite://"
        elif self['engine'] == 'sqlite':
            return f"sqlite:///{self['name']}"
        elif self['engine'] == 'postgresql':
            scheme = 'postgresql+psycopg2'
            if use_async:
                scheme = 'postgresql+asyncpg'
            dsn = f"{scheme}://"
            if self.get('username'):
                dsn += self['username']
                if self.get('password'):
                    dsn += f":{self['password']}"
                dsn += "@"
            dsn += f"{self.get('host') or 'localhost'}"
            dsn += f":{self.get('port') or 5432}"
            if self.get('name'):
                dsn += f"/{self['name']}"
            return dsn
        else:
            raise NotImplementedError # pragma: no cover

    def as_engine(self, create_engine, *args, **kwargs):
        """Return a :class:`sqlalchemy.engine.Engine` instance configured
        with the connection parameters.
        """
        return create_engine(self.as_dsn(kwargs.pop('use_async', False)),
            *args, **kwargs)


def load_config(*args, **kwargs):
    """Like :func:`~unimatrix.lib.rdbms.load_config`, but returns a mapping
    of :class:`DatabaseConnection` objects.
    """
    connections = unimatrix.lib.rdbms.load_config(*args, **kwargs)
    for name, opts in dict.items(connections):
        connections[name] = DatabaseConnection(**opts)
    return connections


DATABASES = load_config(env=copy.deepcopy(os.environ))



"""Declares various functions to assist in testing with :mod:`sqlalchemy`."""
import copy
import os
import unittest

from ioc.loader import import_symbol
from sqlalchemy import create_engine
from sqlalchemy import event
from sqlalchemy.orm import sessionmaker

from unimatrix.ext.orm import Relation
from unimatrix.ext.orm.conf import load_config


class TestCase(unittest.TestCase):
    """A :class:`unittest.TestCase` implementation that ensures a clean
    database state between tests.
    """
    default_bases = [Relation]

    #: Additional base classes for SQLAlchemy models.
    bases = None

    @classmethod
    def setUpClass(cls):
        cls.manager = TestDatabaseManager(
            self.default_bases + list(self.bases or []))
        cls.manager.on_setup_class({
            'DB_ENGINE': 'sqlite',
            'DB_NAME': ":memory:"
        })

    def setUp(self):
        """Ensures that all connections, transactions and sessions are
        started.
        """
        self.manager.on_setup(self)

    def tearDown(self):
        """Rolls back all transactions and closes the connections."""
        self.manager.on_teardown()


class TestDatabaseManager:
    """Manages the state of the database(s) during automated tests."""

    def __init__(self, bases):
        self.bases = bases
        self.databases = {}
        self.engines = {}
        self.connections = {}
        self.transactions = {}
        self.sessions = {}
        self.session = None

    def get_session(self, name='self'):
        """Return a session for the named connection."""
        return sessionmaker(bind=self.engines[name])()

    def on_setup_class(self, databases=None):
        """Prepare the database connections to commence testing."""
        # Get the database from the environment and secrets directory, and
        # configure a default connection if one is not specified.
        self.databases = load_config(env=copy.deepcopy(os.environ))
        if not self.databases:
            self.databases = load_config(env={
                'DB_ENGINE': "sqlite",
                'DB_NAME': ":memory:"
            })

        # Loop over the discovered database connections to create the engine
        # objects.
        for name, opts in dict.items(self.databases):
            self.engines[name] = engine = opts.as_engine(create_engine, echo=True)
            if name == 'self':
                for Base in self.bases:
                    Base.metadata.create_all(engine)

    def on_setup(self, testcase):
        """Ensures that all sessions for the configured databases have started
        a transaction.
        """

        # TODO: Use the default engine to create the tables. Multi-database
        # support is not fully implemented like Django, but this is good
        # enough for now.
        for name in dict.keys(self.databases):
            engine = self.engines[name]
            self.connections[name] = connection = engine.connect()
            self.transactions[name] = tx = connection.begin()
            self.sessions[name] = session = sessionmaker(bind=connection)()
            if name == 'self':
                testcase.session = session
            #session.begin_nested()

            #@event.listens_for(session, "after_transaction_end")
            #def restart_savepoint(session, tx):
            #    if tx.nested and not tx._parent.nested:
            #        session.begin_nested()

    def on_teardown(self):
        """Rollback all transactions and dispose the engines."""
        for session in dict.values(self.sessions):
            session.close()
        for tx in dict.values(self.transactions):
            tx.rollback()
        for connection in dict.values(self.connections):
            connection.close()



# pylint: skip-file
import unittest

from sqlalchemy import create_engine

from unimatrix.ext.orm import Relation


class RelationBasicTestCase(unittest.TestCase):

    def setUp(self):
        self.engine = create_engine('sqlite://')

    def tearDown(self):
        self.engine.dispose()

    def test_create_all(self):
        Relation.metadata.create_all(self.engine)



# pylint: skip-file
import unittest

from unimatrix.ext.orm.conf import load_config


class DatabaseConfigurationTestCase(unittest.TestCase):

    def test_env_sqlite_memory(self):
        env = {
            'DB_ENGINE': "sqlite",
            'DB_NAME': ":memory:"
        }
        connections = load_config(env)
        self.assertEqual(connections['self'].as_dsn(), "sqlite://")

    def test_env_sqlite_disk_unix_absolute(self):
        env = {
            'DB_ENGINE': "sqlite",
            'DB_NAME': "/tmp/foo.sqlite3"
        }
        connections = load_config(env)
        self.assertEqual(connections['self'].as_dsn(), "sqlite:////tmp/foo.sqlite3")

    def test_env_sqlite_disk_unix_relative(self):
        env = {
            'DB_ENGINE': "sqlite",
            'DB_NAME': "tmp/foo.sqlite3"
        }
        connections = load_config(env)
        self.assertEqual(connections['self'].as_dsn(), "sqlite:///tmp/foo.sqlite3")



# pylint: skip-file
import asyncio
import unittest

from .. import create_engine
from .. import destroy_engines
from .. import async_session
from .. import session_factory
from ..conf import load_config


class SessionTestCase(unittest.TestCase):

    def setUp(self):
        self.engine = create_engine(
            databases=load_config({
                'DB_ENGINE': 'sqlite',
                'DB_NAME': ":memory:"
            })
        )

    def tearDown(self):
        destroy_engines()

    def test_create_session_with_no_arguments(self):
        Session = session_factory()
        self.assertEqual(Session().bind.dialect.name, 'sqlite')

    def test_create_session_with_no_arguments(self):
        async def f():
            async with async_session() as session:
                pass

        asyncio.run(f())
